import importlib
import inspect
from collections.abc import MutableMapping
from copy import deepcopy
from typing import Any, Optional, Union

__all__ = ["initialize", "import_object"]


def import_object(object_path: str, reload_module=False):
    """
    Dynamically import a python object(e.g. class, function, ...) from a string reference if it's location.

    - `object_path`: The path to the Python object to import.
    Example: `"functools.lru_cache"`, `"torchvision.transforms.Resize"`
    """

    assert "." in object_path
    module, class_name = object_path.rsplit(".", 1)

    try:
        module_imp = importlib.import_module(module)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(f"failed to import `{object_path}`: {e}") from None

    if reload_module:
        importlib.reload(module_imp)
    try:
        cls = getattr(module_imp, class_name)
    except AttributeError as e:
        raise AssertionError(f"import module `{module}` over, but failed import class `{class_name}`: {e}") from None
    return cls


def _initialize(obj_type, arguments, initialize_classmethod: Optional[str] = None) -> Any:
    if isinstance(obj_type, str):
        obj_cls = import_object(obj_type)
    elif inspect.isclass(obj_type):
        obj_cls = obj_type
    else:
        raise TypeError(f"`obj_type` must be a class name of valid class, " f"but got {type(obj_type)}")

    try:
        if initialize_classmethod is None:
            obj = obj_cls(**arguments)
        else:
            init_method = getattr(obj_cls, initialize_classmethod)
            obj = init_method(**arguments)
    except Exception as e:
        raise TypeError(f"Invalid arguments in {arguments} while " f"building class {obj_cls}") from e

    return obj


def initialize(
    config: Union[str, MutableMapping],
    default_arguments: Optional[MutableMapping] = None,
    initialize_classmethod: Optional[str] = None,
) -> Any:
    """
    initialize a class instance with `config`.

    - `config` can be as class "location.name" str or a Mapping.
    `import_object` will be used to import desired class.

    - `default_arguments` will be used as the kwargs to initialize this class.

    Example:

    1. `config` as a str: `torchvision.models.resnet50`
    2. `config` as a dict: `dict(_type="torchvision.models.resnet50", pretrained=True)` or
      `{"torchvision.models.resnet50": dict(pretrained=True)}`

    `default_arguments` can be used to provide public arguments no matter what `config` you use:

    ```python
    raydl.initialize(config="torch.optim.SGD", default_arguments=dict(lr=0.0001))
    raydl.initialize(config="torch.optim.Adam", default_arguments=dict(lr=0.0001))
    ```

    """
    # get `obj_type``, `arguments`
    if isinstance(config, MutableMapping):
        if "_type" in config:
            arguments = deepcopy(config)
            obj_type = arguments.pop("_type")
            arguments = dict(arguments.items())
        elif len(config) == 1:
            obj_type, arguments = tuple(config.items())[0]
            arguments = dict(arguments.items())
        else:
            raise ValueError(
                "Invalid `config`, Mapping `config` must contain "
                "the type information by dict(_type=CLASS_NAME, ...arguments) "
                f"or dict(CLASS_NAME=dict(...arguments)), but got {config}"
            )
    elif isinstance(config, str):
        obj_type, arguments = config, {}
    else:
        raise TypeError(
            f"`config` must be a str or a `MutableMapping` like dict," f" but got {config} with type {type(config)}"
        )

    if not (isinstance(default_arguments, MutableMapping) or default_arguments is None):
        raise TypeError("default_args must be a dict or None, " f"but got {type(default_arguments)}")
    if default_arguments is not None:
        for k, v in default_arguments.items():
            arguments.setdefault(k, v)

    return _initialize(obj_type, arguments, initialize_classmethod)



from .collect_env import collect_env
from .initialization import import_object, initialize
from .tensor import slerp
from .time import Timer, running_timer
from .utils import manual_seed

__version__ = "0.0.3"



from collections.abc import Callable
from contextlib import contextmanager
from time import perf_counter
from typing import Any, Optional

__all__ = ["Timer", "running_timer"]


class Timer:
    def __init__(self, average: bool = False):
        self._average = average

        self.reset()

    def reset(self, *args: Any) -> "Timer":
        """Reset the timer to zero."""
        self._t0 = perf_counter()
        self.total = 0.0
        self.step_count = 0.0
        self.running = True

        self._last_check_time = 0.0
        self._last_step_time = 0.0

        self._last_step_value = 0.0
        self._last_step_elapsed = 0.0

        return self

    def pause(self, *args: Any) -> None:
        """Pause the current running timer."""
        if self.running:
            self.total += self._elapsed()
            self.running = False

    def resume(self, *args: Any) -> None:
        """Resume the current running timer."""
        if not self.running:
            self.running = True
            self._t0 = perf_counter()

    def since_start(self, do_not_update_check_time=False) -> float:
        total = self.total
        if self.running:
            total += self._elapsed()
        if not do_not_update_check_time:
            self._last_check_time = total
        return total

    def since_last_check(self) -> float:
        last_check_time = self._last_check_time
        return self.since_start() - last_check_time

    def value(self) -> float:
        total = self.since_start()
        denominator = max(self.step_count, 1.0) if self._average else 1.0
        return total / denominator

    def step(self, *args: Any) -> None:
        """Increment the timer."""
        self.step_count += 1.0

        total = self.since_start(do_not_update_check_time=True)
        self._last_step_elapsed = total - self._last_step_value
        self._last_step_value = total

    def _elapsed(self) -> float:
        return perf_counter() - self._t0


@contextmanager
def running_timer(echo_func: Optional[Callable] = lambda x: print(f"{x:.4f}"), average=False):
    timer = Timer(average=average)
    try:
        yield timer
    finally:
        if echo_func is not None:
            echo_func(timer.value())



import contextlib
import random

import torch

__all__ = ["manual_seed"]


def manual_seed(seed: int) -> None:
    """Setup random state from a seed for `torch`, `random` and optionally `numpy` (if can be imported).
    Args:
        seed: Random state seed
    """
    random.seed(seed)
    torch.manual_seed(seed)

    with contextlib.suppress(ImportError):
        import numpy as np

        np.random.seed(seed)



import time

import torch.cuda
import torch.distributed as dist
import torch.utils.collect_env

__all__ = ["collect_env"]

# Composer environment information string output format
_DISTRIBUTED_CONFIGURE_FORMAT = """
Backend: {backend}
World size: {world_size}
Rank: {rank}
""".strip()


def collect_distributed_configure() -> dict:
    if not dist.is_initialized():
        return dict(backend=None, world_size=1, rank=0)

    return {
        "backend": dist.get_backend(),
        "world_size": dist.get_world_size(),
        "rank": dist.get_rank(),
    }


def collect_running_env() -> str:
    out = "local GPU count: {cuda_device_count}\n".format(
        cuda_device_count=torch.cuda.device_count(),
    )

    out += _DISTRIBUTED_CONFIGURE_FORMAT.format(**collect_distributed_configure())

    return out


def collect_env() -> str:
    # Creation timestamp for report
    creation_time = time.strftime("%Y-%m-%d %H:%M:%S %Z", time.localtime(time.time()))
    report_hdr = (
        "---------------------------------\n"
        + "System Environment Report        \n"
        + f"Created: {creation_time}\n"
        + "---------------------------------\n"
    )

    # Torch section
    out = report_hdr
    out += "\nPyTorch information\n"
    out += "-------------------\n"
    out += torch.utils.collect_env.get_pretty_env_info()

    out += "\n\nRunning information\n"
    out += "--------------------\n"
    out += collect_running_env()

    return out



from typing import Union

from torch import FloatTensor, Size, Tensor, lerp, zeros_like
from torch.linalg import norm


def slerp(v0: FloatTensor, v1: FloatTensor, t: Union[float, FloatTensor], DOT_THRESHOLD=0.9995):
    """
    Spherical linear interpolation
    Args:
        v0: Starting vector
        v1: Final vector
        t: Float value between 0.0 and 1.0 or many-dimensional tensor
        DOT_THRESHOLD: Threshold for considering the two vectors as
                               colinear. Not recommended to alter this.
    Returns:
        Interpolation vector between v0 and v1

    - many-dimensional vectors
    - v0 or v1 with last dim all zeroes, or v0 ~colinear with v1
        - falls back to lerp()
        - conditional logic implemented with parallelism rather than Python loops
    - many-dimensional tensor for t
        - you can ask for batches of slerp outputs by making t more-dimensional than the vectors
        -   slerp(
                v0:   torch.Size([2,3]),
                v1:   torch.Size([2,3]),
                t:  torch.Size([4,1,1]),
            )
        - this makes it interface-compatible with lerp()

    see more example at https://gist.github.com/Birch-san/230ac46f99ec411ed5907b0a3d728efa
    """
    assert v0.shape == v1.shape, "shapes of v0 and v1 must match"

    # Normalize the vectors to get the directions and angles
    v0_norm: FloatTensor = norm(v0, dim=-1)
    v1_norm: FloatTensor = norm(v1, dim=-1)

    v0_normed: Tensor = v0 / v0_norm.unsqueeze(-1)
    v1_normed: Tensor = v1 / v1_norm.unsqueeze(-1)

    # Dot product with the normalized vectors
    dot: Tensor = (v0_normed * v1_normed).sum(-1)
    dot_mag: Tensor = dot.abs()

    # if dp is NaN, it's because the v0 or v1 row was filled with 0s
    # If absolute value of dot product is almost 1, vectors are ~colinear, so use lerp
    gotta_lerp = dot_mag.isnan() | (dot_mag > DOT_THRESHOLD)
    can_slerp = ~gotta_lerp

    t_batch_dim_count: int = max(0, t.dim() - v0.dim()) if isinstance(t, Tensor) else 0
    t_batch_dims: Size = t.shape[:t_batch_dim_count] if isinstance(t, Tensor) else Size([])
    out = zeros_like(v0.expand(*t_batch_dims, *[-1] * v0.dim()))

    # if no elements are lerpable, our vectors become 0-dimensional, preventing broadcasting
    if gotta_lerp.any():
        v0_l = v0.masked_select(gotta_lerp.unsqueeze(-1)).unflatten(
            dim=-1, sizes=(-1, *[1] * (v0.dim() - 2), v0.size(-1))
        )
        v1_l = v1.masked_select(gotta_lerp.unsqueeze(-1)).unflatten(
            dim=-1, sizes=(-1, *[1] * (v1.dim() - 2), v1.size(-1))
        )

        lerped = lerp(v0_l, v1_l, t)

        out = out.masked_scatter_(gotta_lerp.unsqueeze(-1).expand(*t_batch_dims, *gotta_lerp.shape, 1), lerped)

    # if no elements are slerpable, our vectors become 0-dimensional, preventing broadcasting
    if can_slerp.any():
        v0_s = v0.masked_select(can_slerp.unsqueeze(-1)).unflatten(
            dim=-1, sizes=(-1, *[1] * (v0.dim() - 2), v0.size(-1))
        )
        v1_s = v1.masked_select(can_slerp.unsqueeze(-1)).unflatten(
            dim=-1, sizes=(-1, *[1] * (v1.dim() - 2), v1.size(-1))
        )
        dot_s = dot.masked_select(can_slerp).unflatten(dim=-1, sizes=(-1, *[1] * (v1.dim() - 1)))

        # Calculate initial angle between v0 and v1
        theta_0: Tensor = dot_s.arccos()
        sin_theta_0: Tensor = theta_0.sin()
        # Angle at timestep t
        theta_t: Tensor = theta_0 * t
        sin_theta_t: Tensor = theta_t.sin()
        # Finish the slerp algorithm
        s0: Tensor = (theta_0 - theta_t).sin() / sin_theta_0
        s1: Tensor = sin_theta_t / sin_theta_0
        slerped: Tensor = s0 * v0_s + s1 * v1_s

        return out.masked_scatter_(can_slerp.unsqueeze(-1).expand(*t_batch_dims, *can_slerp.shape, 1), slerped)

    return out
